@@ -35,6 +35,9 @@
 
 #include "near.h"
 
+/* We check for the tag being present every 2 seconds */
+#define CHECK_PRESENCE_PERIOD 2
+
 static DBusConnection *connection = NULL;
 
 static GHashTable *adapter_hash;
@@ -48,6 +51,8 @@ struct near_adapter {
 
 	near_bool_t powered;
 	near_bool_t polling;
+	near_bool_t constant_poll;
+	near_bool_t dep_up;
 
 	GHashTable *targets;
 	struct near_target *link;
@@ -57,6 +62,8 @@ struct near_adapter {
 	guint watch;
 	GList *ioreq_list;
 	GList *ndef_q;
+
+	guint presence_timeout;
 };
 
 struct near_adapter_ioreq {
@@ -76,6 +83,9 @@ static void free_adapter(gpointer data)
 {
 	struct near_adapter *adapter = data;
 
+	if (adapter->presence_timeout > 0)
+		g_source_remove(adapter->presence_timeout);
+
 	g_free(adapter->name);
 	g_free(adapter->path);
 	g_free(adapter);
@@ -96,6 +106,28 @@ static void polling_changed(struct near_adapter *adapter)
 					DBUS_TYPE_BOOLEAN, &adapter->polling);
 }
 
+static int adapter_start_poll(struct near_adapter *adapter)
+{
+	int err;
+
+	if (g_hash_table_size(adapter->targets) > 0) {
+		DBG("Clearing targets");
+
+		g_hash_table_remove_all(adapter->targets);
+		__near_adapter_target_changed(adapter->idx);
+	}
+
+	err = __near_netlink_start_poll(adapter->idx, adapter->protocols);
+	if (err < 0)
+		return err;
+
+	adapter->polling = TRUE;
+
+	polling_changed(adapter);
+
+	return 0;
+}
+
 static void append_path(gpointer key, gpointer value, gpointer user_data)
 {
 	struct near_adapter *adapter = value;
@@ -198,7 +230,6 @@ void __near_adapter_target_changed(uint32_t adapter_idx)
 				DBUS_TYPE_OBJECT_PATH, append_targets,
 				adapter);
 }
-
 static DBusMessage *get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -236,8 +267,38 @@ static DBusMessage *get_properties(DBusConnection *conn,
 static DBusMessage *set_property(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
+	struct near_adapter *adapter = data;
+	DBusMessageIter iter, value;
+	const char *name;
+	int type, err;
+
 	DBG("conn %p", conn);
 
+	if (dbus_message_iter_init(msg, &iter) == FALSE)
+		return __near_error_invalid_arguments(msg);
+
+	dbus_message_iter_get_basic(&iter, &name);
+	dbus_message_iter_next(&iter);
+	dbus_message_iter_recurse(&iter, &value);
+
+	type = dbus_message_iter_get_arg_type(&value);
+
+	if (g_str_equal(name, "Powered") == TRUE) {
+		near_bool_t powered;
+
+		if (type != DBUS_TYPE_BOOLEAN)
+			return __near_error_invalid_arguments(msg);
+
+		dbus_message_iter_get_basic(&value, &powered);
+
+		err = __near_netlink_adapter_enable(adapter->idx, powered);
+		if (err < 0)
+			return __near_error_failed(msg, -err);
+
+		adapter->powered = powered;
+	} else
+		return __near_error_invalid_property(msg);
+
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
@@ -249,21 +310,10 @@ static DBusMessage *start_poll(DBusConnection *conn,
 
 	DBG("conn %p", conn);
 
-	if (g_hash_table_size(adapter->targets) > 0) {
-		DBG("Clearing targets");
-
-		g_hash_table_remove_all(adapter->targets);
-		__near_adapter_target_changed(adapter->idx);
-	}
-
-	err = __near_netlink_start_poll(adapter->idx, adapter->protocols);
+	err = adapter_start_poll(adapter);
 	if (err < 0)
 		return __near_error_failed(msg, -err);
 
-	adapter->polling = TRUE;
-
-	polling_changed(adapter);
-
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
@@ -361,10 +411,189 @@ static int __publish_text_record(DBusMessage *msg, void *data)
 	return __push_ndef_queue(data, ndef);
 }
 
+static void tag_present_cb(uint32_t adapter_idx, uint32_t target_idx,
+								int status);
+
+static gboolean check_presence(gpointer user_data)
+{
+	struct near_adapter *adapter = user_data;
+	struct near_target *target;
+	int err;
+
+	DBG("");
+
+	if (adapter == NULL)
+		return FALSE;
+
+	target = adapter->link;
+	if (target == NULL)
+		return FALSE;
+
+	err = __near_tag_check_presence(target, tag_present_cb);
+	if (err < 0) {
+		DBG("Could not check target presence");
+
+		near_adapter_disconnect(adapter->idx);
+		if (adapter->constant_poll == TRUE)
+			adapter_start_poll(adapter);
+	}
+
+	return FALSE;
+}
+
+static void tag_present_cb(uint32_t adapter_idx, uint32_t target_idx,
+								int status)
+{
+	struct near_adapter *adapter;
+
+	DBG("");
+
+	adapter = g_hash_table_lookup(adapter_hash,
+					GINT_TO_POINTER(adapter_idx));
+	if (adapter == NULL)
+		return;
+
+	if (status < 0) {
+		DBG("Target is gone");
+
+		near_adapter_disconnect(adapter->idx);
+		if (adapter->constant_poll == TRUE)
+			adapter_start_poll(adapter);
+
+		return;
+	}
+
+	adapter->presence_timeout =
+		g_timeout_add_seconds(CHECK_PRESENCE_PERIOD,
+					check_presence, adapter);
+}
+
+static char *get_uri_field(DBusMessage *msg)
+{
+	DBusMessageIter iter, arr_iter;
+	char *uri = NULL;
+
+	DBG("");
+
+	dbus_message_iter_init(msg, &iter);
+	dbus_message_iter_recurse(&iter, &arr_iter);
+
+	while (dbus_message_iter_get_arg_type(&arr_iter) !=
+					DBUS_TYPE_INVALID) {
+		const char *key;
+		DBusMessageIter ent_iter;
+		DBusMessageIter var_iter;
+
+		dbus_message_iter_recurse(&arr_iter, &ent_iter);
+		dbus_message_iter_get_basic(&ent_iter, &key);
+		dbus_message_iter_next(&ent_iter);
+		dbus_message_iter_recurse(&ent_iter, &var_iter);
+
+		switch (dbus_message_iter_get_arg_type(&var_iter)) {
+		case DBUS_TYPE_STRING:
+			if (g_strcmp0(key, "URI") == 0)
+				dbus_message_iter_get_basic(&var_iter, &uri);
+
+			break;
+		}
+
+		dbus_message_iter_next(&arr_iter);
+	}
+
+	return uri;
+}
+
+static int publish_uri_record(DBusMessage *msg, void *data)
+{
+	struct near_ndef_message *ndef;
+	char *uri = NULL;
+	const char *uri_prefix = NULL;
+	uint8_t id_len, i;
+	uint32_t uri_len;
+
+	DBG("");
+
+	uri = get_uri_field(msg);
+	if (uri == NULL)
+		return -EINVAL;
+
+	for (i = 1; i <= NFC_MAX_URI_ID; i++) {
+		uri_prefix = __near_ndef_get_uri_prefix(i);
+
+		if (uri_prefix != NULL &&
+				g_str_has_prefix(uri, uri_prefix) == TRUE)
+			break;
+	}
+
+	/* If uri_prefix is NULL then ID will be zero */
+	if (uri_prefix == NULL) {
+		i = 0;
+		id_len = 0;
+	} else
+		id_len = strlen(uri_prefix);
+
+	uri_len = strlen(uri) - id_len;
+	ndef = near_ndef_prepare_uri_record(i, uri_len,
+						(uint8_t *)(uri + id_len));
+	if (ndef == NULL)
+		return -EINVAL;
+
+	return __push_ndef_queue(data, ndef);
+}
+
+static int publish_sp_record(DBusMessage *msg, void *data)
+{
+	struct near_ndef_message *ndef;
+	char *uri = NULL;
+	const char *uri_prefix;
+	uint8_t id_len, i;
+	uint32_t uri_len;
+
+	DBG("");
+
+	/* Currently this funtion supports only mandatory URI record,
+	 * TODO: Other records support */
+	uri = get_uri_field(msg);
+	if (uri == NULL)
+		return -EINVAL;
+
+	for (i = 1; i <= NFC_MAX_URI_ID; i++) {
+		uri_prefix = __near_ndef_get_uri_prefix(i);
+
+		if (uri_prefix != NULL &&
+				g_str_has_prefix(uri, uri_prefix) == TRUE)
+			break;
+	}
+
+	if (uri_prefix == NULL) {
+		i = 0;
+		id_len = 0;
+	} else
+		id_len = strlen(uri_prefix);
+
+	uri_len = strlen(uri) - id_len;
+	ndef = near_ndef_prepare_smartposter_record(i, uri_len,
+						(uint8_t *)(uri + id_len));
+	if (ndef == NULL)
+		return -EINVAL;
+
+	return __push_ndef_queue(data, ndef);
+}
+
 static void __add_ndef_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 {
-	DBG(" %d", status);
-	near_adapter_disconnect(adapter_idx);
+	struct near_adapter *adapter;
+
+	DBG("%d", status);
+
+	adapter = g_hash_table_lookup(adapter_hash,
+					GINT_TO_POINTER(adapter_idx));
+	if (adapter == NULL)
+		return;
+
+	adapter->presence_timeout =
+		g_timeout_add_seconds(CHECK_PRESENCE_PERIOD,
+					check_presence, adapter);
 }
 
 static DBusMessage *publish(DBusConnection *conn,
@@ -388,27 +617,41 @@ static DBusMessage *publish(DBusConnection *conn,
 		dbus_message_iter_recurse(&arr_iter, &ent_iter);
 		dbus_message_iter_get_basic(&ent_iter, &key);
 
-		if (g_strcmp0(key, "Type") == 0) {
-			dbus_message_iter_next(&ent_iter);
-			dbus_message_iter_recurse(&ent_iter, &var_iter);
+		if (g_strcmp0(key, "Type") != 0) {
+			dbus_message_iter_next(&arr_iter);
+			continue;
+		}
 
-			switch (dbus_message_iter_get_arg_type(&var_iter)) {
-			case DBUS_TYPE_STRING:
-				dbus_message_iter_get_basic(&var_iter, &value);
+		dbus_message_iter_next(&ent_iter);
+		dbus_message_iter_recurse(&ent_iter, &var_iter);
 
-				if (g_strcmp0(value, "Text") == 0) {
-					if (__publish_text_record(msg, adapter)
-							< 0)
+		switch (dbus_message_iter_get_arg_type(&var_iter)) {
+		case DBUS_TYPE_STRING:
+			dbus_message_iter_get_basic(&var_iter, &value);
+
+			if (g_strcmp0(value, "Text") == 0) {
+				if (__publish_text_record(msg, adapter)	< 0)
 						goto error;
 
-					goto reply;
-				} else {
-					DBG(" '%s' not supported", value);
+				goto reply;
+			} else if (g_strcmp0(value, "URI") == 0) {
+				if (publish_uri_record(msg, adapter)	< 0)
 					goto error;
-				}
+
+				goto reply;
+
+			} else if (g_strcmp0(value, "SmartPoster") == 0) {
+				if (publish_sp_record(msg, adapter) < 0)
+					goto error;
+
+				goto reply;
+
+			} else {
+				DBG(" '%s' not supported", value);
+				goto error;
+			}
 
 				break;
-			}
 		}
 
 		dbus_message_iter_next(&arr_iter);
@@ -439,7 +682,7 @@ static GDBusSignalTable adapter_signals[] = {
 };
 
 struct near_adapter * __near_adapter_create(uint32_t idx,
-					const char *name, uint32_t protocols)
+		const char *name, uint32_t protocols, near_bool_t powered)
 {
 	struct near_adapter *adapter;
 
@@ -454,7 +697,9 @@ struct near_adapter * __near_adapter_create(uint32_t idx,
 	}
 	adapter->idx = idx;
 	adapter->protocols = protocols;
-	adapter->powered = TRUE;
+	adapter->powered = powered;
+	adapter->constant_poll = near_setting_get_bool("ConstantPoll");
+	adapter->dep_up = FALSE;
 	adapter->targets = g_hash_table_new_full(g_direct_hash, g_direct_equal,
 							NULL, free_target);
 	adapter->sock = -1;
@@ -481,6 +726,34 @@ struct near_adapter *__near_adapter_get(uint32_t idx)
 	return g_hash_table_lookup(adapter_hash, GINT_TO_POINTER(idx));
 }
 
+int near_adapter_get_dep_state(uint32_t idx)
+{
+	struct near_adapter *adapter;
+
+	DBG("idx %d", idx);
+
+	adapter = g_hash_table_lookup(adapter_hash, GINT_TO_POINTER(idx));
+	if (adapter == NULL)
+		return -ENODEV;
+
+	return adapter->dep_up;
+}
+
+int near_adapter_set_dep_state(uint32_t idx, near_bool_t dep)
+{
+	struct near_adapter *adapter;
+
+	DBG("idx %d", idx);
+
+	adapter = g_hash_table_lookup(adapter_hash, GINT_TO_POINTER(idx));
+	if (adapter == NULL)
+		return -ENODEV;
+
+	adapter->dep_up = dep;
+
+	return 0;
+}
+
 int __near_adapter_add(struct near_adapter *adapter)
 {
 	uint32_t idx = adapter->idx;
@@ -518,6 +791,11 @@ static int dep_link_up(uint32_t idx, uint32_t target_idx)
 					NFC_COMM_ACTIVE, NFC_RF_INITIATOR);
 }
 
+static int dep_link_down(uint32_t idx)
+{
+	return __near_netlink_dep_link_down(idx);
+}
+
 static void tag_read_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 {
 	struct near_adapter *adapter;
@@ -526,18 +804,25 @@ static void tag_read_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 	uint16_t tag_type;
 	int err;
 
-	if (status < 0)
-		goto out;
+	DBG("");
 
-	__near_adapter_target_changed(adapter_idx);
-
-	/* Check if adapter ndef queue has any ndef messages,
-	 * then write the ndef data on tag. */
 	adapter = g_hash_table_lookup(adapter_hash,
 					GINT_TO_POINTER(adapter_idx));
 	if (adapter == NULL)
 		goto out;
 
+	if (status < 0) {
+		near_adapter_disconnect(adapter->idx);
+		if (adapter->constant_poll == TRUE)
+			adapter_start_poll(adapter);
+
+		return;
+	}
+
+	__near_adapter_target_changed(adapter_idx);
+
+	/* Check if adapter ndef queue has any ndef messages,
+	 * then write the ndef data on tag. */
 	if (g_list_length(adapter->ndef_q) == 0)
 		goto out;
 
@@ -553,8 +838,9 @@ static void tag_read_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 	tag_type = __near_target_get_tag_type(target);
 
 	/* Add NDEF header information depends upon tag type */
-	if (tag_type & NEAR_TAG_NFC_TYPE1 ||
-			tag_type & NEAR_TAG_NFC_TYPE2) {
+	switch (tag_type) {
+	case NFC_PROTO_JEWEL:
+	case NFC_PROTO_MIFARE:
 		ndef_with_header = g_try_malloc0(sizeof(
 					struct near_ndef_message));
 		if (ndef_with_header == NULL)
@@ -571,7 +857,26 @@ static void tag_read_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 		memcpy(ndef_with_header->data + 2, ndef->data, ndef->length);
 		ndef_with_header->data[ndef->length + 2] = TLV_END;
 
-	} else if (tag_type & NEAR_TAG_NFC_TYPE4) {
+		break;
+
+	case NFC_PROTO_FELICA:
+		ndef_with_header = g_try_malloc0(sizeof(
+					struct near_ndef_message));
+		if (ndef_with_header == NULL)
+			goto out;
+
+		ndef_with_header->offset = 0;
+		ndef_with_header->length = ndef->length;
+		ndef_with_header->data = g_try_malloc0(
+						ndef_with_header->length);
+		if (ndef_with_header->data == NULL)
+			goto out;
+
+		memcpy(ndef_with_header->data, ndef->data, ndef->length);
+
+		break;
+
+	case NFC_PROTO_ISO14443:
 		ndef_with_header = g_try_malloc0(sizeof(
 					struct near_ndef_message));
 		if (ndef_with_header == NULL)
@@ -586,8 +891,12 @@ static void tag_read_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 		ndef_with_header->data[0] = (uint8_t)(ndef->length >> 8);
 		ndef_with_header->data[1] = (uint8_t)(ndef->length);
 		memcpy(ndef_with_header->data + 2, ndef->data, ndef->length);
-	} else
+
+		break;
+
+	default:
 		goto out;
+	}
 
 	g_free(ndef->data);
 	g_free(ndef);
@@ -607,7 +916,10 @@ out:
 		g_free(ndef_with_header);
 	}
 
-	near_adapter_disconnect(adapter_idx);
+	if (adapter != NULL)
+		adapter->presence_timeout =
+			g_timeout_add_seconds(CHECK_PRESENCE_PERIOD,
+						check_presence, adapter);
 }
 
 int __near_adapter_add_target(uint32_t idx, uint32_t target_idx,
@@ -616,6 +928,8 @@ int __near_adapter_add_target(uint32_t idx, uint32_t target_idx,
 {
 	struct near_adapter *adapter;
 	struct near_target *target;
+	uint16_t tag_type;
+	int err;
 
 	DBG("idx %d", idx);
 
@@ -635,12 +949,28 @@ int __near_adapter_add_target(uint32_t idx, uint32_t target_idx,
 	g_hash_table_insert(adapter->targets,
 			GINT_TO_POINTER(target_idx), target);	
 
-	__near_tag_read(target, tag_read_cb);
+	tag_type = __near_target_get_tag_type(target);
 
-	if (protocols & NFC_PROTO_NFC_DEP_MASK)
-		dep_link_up(idx, target_idx);
+	if (tag_type != NFC_PROTO_NFC_DEP) {
+		err = near_adapter_connect(idx, target_idx, tag_type);
+		if (err < 0) {
+			near_error("Could not connect");
+			return err;
+		}
 
-	return 0;
+		err = __near_tag_read(target, tag_read_cb);
+	} else {
+		/* For p2p, reading is service binding */
+		err = __near_tag_read(target, tag_read_cb);
+		if (err < 0) {
+			near_error("Could not read tag");
+			return err;
+		}
+
+		err = near_adapter_connect(idx, target_idx, tag_type);
+	}
+
+	return err;
 }
 
 int __near_adapter_remove_target(uint32_t idx, uint32_t target_idx)
@@ -705,7 +1035,6 @@ static gboolean adapter_recv_event(GIOChannel *channel, GIOCondition condition,
 		near_error("Error while reading NFC bytes");
 
 		adapter_flush_rx(adapter, -EIO);
-		near_adapter_disconnect(adapter->idx);
 		return FALSE;
 	}
 
@@ -745,6 +1074,18 @@ int near_adapter_connect(uint32_t idx, uint32_t target_idx, uint8_t protocol)
 	if (target == NULL)
 		return -ENOLINK;
 
+	if (protocol == NFC_PROTO_NFC_DEP) {
+		err = dep_link_up(idx, target_idx);
+		if (err < 0)
+			return err;
+
+		adapter->link = target;
+
+		DBG("link %p", adapter->link);
+
+		return 0;
+	}
+
 	sock = socket(AF_NFC, SOCK_SEQPACKET, NFC_SOCKPROTO_RAW);
 	if (sock == -1)
 		return sock;
@@ -780,6 +1121,8 @@ int near_adapter_connect(uint32_t idx, uint32_t target_idx, uint8_t protocol)
 int near_adapter_disconnect(uint32_t idx)
 {
 	struct near_adapter *adapter;
+	uint32_t target_idx;
+	uint16_t tag_type;
 
 	DBG("idx %d", idx);
 
@@ -787,6 +1130,23 @@ int near_adapter_disconnect(uint32_t idx)
 	if (adapter == NULL)
 		return -ENODEV;
 
+	DBG("link %p", adapter->link);
+
+	if (adapter->link == NULL)
+		return -ENOLINK;
+
+	tag_type = __near_target_get_tag_type(adapter->link);
+	target_idx = __near_target_get_idx(adapter->link);
+
+	DBG("tag type %d", tag_type);
+
+	__near_adapter_remove_target(adapter->idx, target_idx);
+
+	if (tag_type == NFC_PROTO_NFC_DEP) {
+		adapter->link = NULL;
+		return dep_link_down(idx);
+	}
+
 	if (adapter->sock == -1)
 		return -ENOLINK;
 
@@ -799,7 +1159,6 @@ int near_adapter_disconnect(uint32_t idx)
 	close(adapter->sock);
 	adapter->sock = -1;
 	adapter->link = NULL;
-	adapter_flush_rx(adapter, -ENOLINK);
 
 	return 0;
 }
